/* Public domain */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "sudoku.h"

unsigned char sudoku_conditions[NB_CONDITIONS][9] = {
	{ 0, 1, 2, 3, 4, 5, 6, 7, 8},
	{ 9, 10, 11, 12, 13, 14, 15, 16, 17},
	{ 18, 19, 20, 21, 22, 23, 24, 25, 26},
	{ 27, 28, 29, 30, 31, 32, 33, 34, 35},
	{ 36, 37, 38, 39, 40, 41, 42, 43, 44},
	{ 45, 46, 47, 48, 49, 50, 51, 52, 53},
	{ 54, 55, 56, 57, 58, 59, 60, 61, 62},
	{ 63, 64, 65, 66, 67, 68, 69, 70, 71},
	{ 72, 73, 74, 75, 76, 77, 78, 79, 80},
	{ 0, 9, 18, 27, 36, 45, 54, 63, 72},
	{ 1, 10, 19, 28, 37, 46, 55, 64, 73},
	{ 2, 11, 20, 29, 38, 47, 56, 65, 74},
	{ 3, 12, 21, 30, 39, 48, 57, 66, 75},
	{ 4, 13, 22, 31, 40, 49, 58, 67, 76},
	{ 5, 14, 23, 32, 41, 50, 59, 68, 77},
	{ 6, 15, 24, 33, 42, 51, 60, 69, 78},
	{ 7, 16, 25, 34, 43, 52, 61, 70, 79},
	{ 8, 17, 26, 35, 44, 53, 62, 71, 80},
	{ 0, 1, 2, 9, 10, 11, 18, 19, 20 },
	{ 3, 4, 5, 12, 13, 14, 21, 22, 23 },
	{ 6, 7, 8, 15, 16, 17, 24, 25, 26 },
	{ 27, 28, 29, 36, 37, 38, 45, 46, 47 },
	{ 30, 31, 32, 39, 40, 41, 48, 49, 50 },
	{ 33, 34, 35, 42, 43, 44, 51, 52, 53 },
	{ 54, 55, 56, 63, 64, 65, 72, 73, 74 },
	{ 57, 58, 59, 66, 67, 68, 75, 76, 77 },
	{ 60, 61, 62, 69, 70, 71, 78, 79, 80 }
};


int
sudoku_check_conditions(sudoku_grid_ref grid)
/* -1 if some condition fail */
/*  0 if ok but the grid is not complete */
/*  1 if the grid is filled */
{
	int i, j;
	unsigned char digits[10];
	int unfinished = 0;
	for (i=0;i<NB_CONDITIONS;i++) {
		memset(digits, 0, sizeof(digits));
		for(j=0;j<9;j++) {
			int digit = grid->number[sudoku_conditions[i][j]];
			if (digit == 0)
				unfinished = 1;
			else {
				if (digits[digit]++) {
					return -1;
				}
			}
		}
	}
	return (unfinished ? 0 : 1);
}

void
init_sudoku_grid(sudoku_grid_ref grid)
{
	memset(grid, 0, sizeof(sudoku_grid_s));
}

sudoku_grid_ref
new_sudoku_grid(void)
{
	sudoku_grid_ref p = malloc(sizeof(sudoku_grid_s));
	if (!p) {
		perror("malloc");
		exit(42);
	}
	init_sudoku_grid(p);
	return p;
}

void
duplicate_sudoku_grid(sudoku_grid_ref dest, sudoku_grid_ref src)
{
	memcpy(dest, src, sizeof(sudoku_grid_s));
}

void
sudoku_fill_grid_from_stdin(sudoku_grid_ref grid)
{
	int x = 0;
	int c;
	while (x<81) {
		if ((c=getchar()) == EOF) {
			fprintf(stderr, "sudoku_fill_grid_from_stdin: imcomplete grid\n");
			exit(42);
		}
		if (((c>='0') && (c<='9')) || (c == '.')) {
			grid->number[x++] = (c == '.') ? 0 : (c - '0');
		}
	}
}

void
sudoku_display_grid(sudoku_grid_ref grid)
{
	int row, col, x = 0;
	for(row=0;row<9;row++) {
		for (col=0;col<9;col++) {
			unsigned char n = grid->number[x++];
			putchar(n ? (n+'0') : '.');
			if ((col == 2) || (col == 5)) putchar('|');
		}
		printf("\n");
		if ((row == 2) || (row == 5))
			printf("---+---+---\n");
	}
	printf("\n");
}

int
sudoku_solve(sudoku_grid_ref grid)
{
	int res = sudoku_check_conditions(grid);
	unsigned char n, try;
	sudoku_grid_ref new_grid;

	if (res == -1) return -1;
	if (res == 1) {
		sudoku_display_grid(grid); /* Found a solution */
		return 0;
	}

	/* Still have to find some digits */
	new_grid = new_sudoku_grid();
	for (n=0;grid->number[n];n++);
	res = -1;
	for (try=1;try<=9;try++) {
		duplicate_sudoku_grid(new_grid, grid);
		new_grid->number[n] = try;
		if (sudoku_solve(new_grid) == 0) res = 0;
	}
	free(new_grid);
	return res;
}

